# pylint: disable=too-few-public-methods

from importlib.util import module_from_spec, spec_from_file_location
import logging
from logging.handlers import WatchedFileHandler
import os
from pathlib import Path
import re
import sys
from typing import Any, Dict, Optional

from git import InvalidGitRepositoryError, Repo
from github3 import GitHub, login
from pydantic import BaseSettings
import toml


class CommonSettings(BaseSettings):
    """
    Common settings which applies to both TOML and CLI
    configuration of Hammurabi.

    ``Pillar`` configuration is intentionally not listed since
    it is represented as a string in the TOML configuration, but
    used the parsed variable in the CLI configuration.
    """

    allow_push: bool = True
    dry_run: bool = False
    rule_can_abort: bool = False
    git_branch_name: str = "hammurabi"
    git_base_name: str = "master"
    repository: str = ""
    report_name: Path = Path("hammurabi_report.json")

    class Config:
        """
        BaseSettings' config describing how the settings will be handled.
        The given ``env_prefix`` will make sure that settings can be read from
        environment variables starting with ``HAMMURABI_``.
        """

        env_prefix = "hammurabi_"


class TOMLSettings(CommonSettings):
    """
    TOML Project configuration settings. Most of the fields
    are used to compose other configuration fields like
    ``github_token`` or ``pillar``.
    """

    github_token: str = ""
    log_level: str = "INFO"
    log_path: Optional[Path] = Path("./hammurabi.log")
    log_format: str = logging.BASIC_FORMAT
    pillar_config: Path = Path("pillar.conf.py")
    pillar_name: str = "pillar"


class Settings(CommonSettings):
    """
    CLI related settings which are directly needed for the
    execution.
    """

    pillar: object = None


class Config:
    """
    Simple configuration object which used across Hammurabi.
    The :class:`Config` loads the given ``pyproject.toml`` according
    to PEP-518.

    .. warning::

        When trying to use GitHub based laws without an initialized GitHub
        client (or invalid token), a warning will be thrown at the beginning
        of the execution. In case a PR open is attempted, a ``RuntimeError``
        will be raised
    """

    def __init__(self) -> None:
        try:
            repo = Repo(self.__get_repo_path())
        except InvalidGitRepositoryError as exc:
            logging.error('"%s" is not a git repository', str(exc))
            repo = None

        self.repo: Repo = repo
        self.github: Optional[GitHub] = None
        self.settings: Settings = Settings()

    @staticmethod
    def __setup_logging(project_config: TOMLSettings) -> None:
        """
        Set the logging configuration of the root logger. The root logger is
        intentionally configured and it is not a mistake.

        :param project_config: Parsed TOMLSettings
        :type project_config: :class:`hammurabi.config.TOMLSettings`
        """

        logging.root.setLevel(project_config.log_level)
        formatter = logging.Formatter(project_config.log_format)

        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(formatter)
        logging.root.addHandler(stream_handler)

        if project_config.log_path:
            file_handler = WatchedFileHandler(str(project_config.log_path))
            file_handler.setFormatter(formatter)
            logging.root.addHandler(file_handler)

    @staticmethod
    def __get_repo_path() -> Path:
        """
        Get repository path which is the current working directory.
        :return: Current working directory where Hammurabi is executed
        :rtype: Path
        """

        return Path(".").absolute()

    @staticmethod
    def __load_pyproject_toml(config_file: Path) -> TOMLSettings:
        """
        Load and parse the given ``pyproject.toml`` file.

        :param config_file: Path of the pyproject.toml file
        :type config_file: Path

        :return: Returns the parsed configuration
        :rtype: :class:`hammurabi.config.TOMLSettings`
        """

        return TOMLSettings(
            **toml.load(config_file).get("tool", {}).get("hammurabi", {})
        )

    @staticmethod
    def __load_pillar_config(project_config: TOMLSettings) -> object:
        """
        Load ``pillar`` configuration based on the dotted style path in the
        ``pyproject.toml``, set by ``config`` configuration section.

        :param project_config: Parsed TOMLSettings
        :type project_config: :class:`hammurabi.config.TOMLSettings`

        :return: Returns the initialized :class:`hammurabi.pillar.Pillar`
        :rtype: :class:`pydantic.PyObject`
        """

        # Pillar configuration file
        pillar_config = project_config.pillar_config.expanduser()

        # Load the configuration from pillar config module to runtime
        spec = spec_from_file_location(
            pillar_config.name.replace(".py", ""), os.path.expandvars(pillar_config)
        )

        module = module_from_spec(spec)
        spec.loader.exec_module(module)  # type: ignore

        # Pillar config variable name
        return getattr(module, project_config.pillar_name)

    @staticmethod
    def __get_settings_value(parameter: str, fallback: Any) -> Any:
        """
        Get a specific setting parameter's value. This helper function
        will prefer environment variables over the fallback value, to
        keep the configuration order.

        :param parameter: Name of the setting
        :type parameter: str

        :param fallback: Value set in the configuration file
        :type fallback: Any

        :return: Returns the value read from environment or config file
        :rtype: Any
        """

        prefix = Settings.Config.env_prefix
        return os.environ.get(f"{prefix}{parameter}", fallback)

    def __merge_settings(self, loaded_settings: Dict[str, Any]) -> Settings:
        """
        Merge the configuration parsed from pyproject.toml and set by
        environment variables. To keep the configuration loading order,
        the environment settings will be used over those found in the
        TOML file.

        Config priority:

            1. CLI arguments (set by the CLI)
            2. ENV Variables (handled by pydantic settings)
            3. Config from file (handled by ``Config`` object)
            4. Default config (handled by pydantic settings)

        :param loaded_settings:
        :type loaded_settings: Dict[str, Any]
        """

        merge_result = {}

        for setting, value in loaded_settings.items():
            merge_result[setting] = self.__get_settings_value(setting, value)

        return Settings(**merge_result)

    def __get_fallback_repository(self) -> str:
        """
        Figure out the fallback owner/repository based on the remote url of the git repo.
        :return: Returns the owner/repository pair
        :rtype: str
        """

        repo_url: str = self.repo.remote().url

        if re.match(r"^http(s)?://", repo_url):
            repo = "/".join(repo_url.split("/")[-2:])
        else:
            repo = repo_url.split(":")[-1]

        return repo.replace(".git", "")

    def load(self):
        """
        Handle configuration loading from project toml file and make sure
        the configuration are initialized and merged. Also, make sure that
        logging is set properly. Before loading the configuration, it is a
        requirement to set the ``HAMMURABI_SETTINGS_PATH`` as it will contain
        the path to the ``toml`` file what Hammurabi expects. This is needed
        for cases when the 3rd party rules would like to read the configuration
        of Hammurabi.

        ... note:

            The ``HAMMURABI_SETTINGS_PATH`` environment variable is set by the CLI
            by default, so there is no need to set if no 3rd party rules are used
            or those rules are not loading config.

        :raises: Runtime error if ``HAMMURABI_SETTINGS_PATH`` environment variable is not
                 set or an invalid git repository was given.

        """

        if not self.repo:
            raise RuntimeError(f'"{self.__get_repo_path()}" is not a git repository.')

        settings_path = Path(
            os.path.expandvars(
                os.environ.get("HAMMURABI_SETTINGS_PATH", "pyproject.toml")
            )
        ).expanduser()

        if not settings_path.exists():
            raise RuntimeError(
                f'Environment variable "HAMMURABI_SETTINGS_PATH" ({settings_path}) '
                "does not exists. Please make sure that you set the environment variable "
                "or CLI ``-c/--config`` flag properly. You must either define the"
                "environment variable or use hammurabi as a CLI tool."
            )

        # Hammurabi CLI configuration file
        project_config = self.__load_pyproject_toml(settings_path)

        # Merge settings and make sure we keep config priority
        # Override the default settings by the merged ones
        self.settings = self.__merge_settings(
            {
                "pillar": self.__load_pillar_config(project_config),
                "git_base_name": project_config.git_base_name,
                "git_branch_name": project_config.git_branch_name,
                "dry_run": project_config.dry_run,
                "rule_can_abort": project_config.rule_can_abort,
                "report_name": project_config.report_name,
            }
        )

        if not project_config.repository:
            self.settings.repository = self.__get_fallback_repository()

        # Set after self.settings is set since the following
        # may depend on settings read from environment or config file
        self.github = login(token=project_config.github_token)

        self.__setup_logging(project_config)

        logging.debug('Successfully loaded "%s"', settings_path)
        logging.debug('Successfully loaded "%s"', project_config.pillar_config)

        if not self.github:
            logging.warning("Missing or invalid GitHub token")


config = Config()  # pylint: disable=invalid-name



"""
Attributes module contains file and directory attribute manipulation
rules which can be handy after creating new files or directories or
even when adding execute permissions for a script in the project.
"""

from abc import abstractmethod
import logging
import os
from pathlib import Path
import shutil
from typing import Any, Optional

from hammurabi.rules.common import SinglePathRule


class SingleAttributeRule(SinglePathRule):
    """
    Extend :class:`hammurabi.rules.base.Rule` to handle attributes of a single
    file or directory.
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        new_value: Optional[str] = None,
        **kwargs
    ) -> None:
        self.new_value = self.validate(new_value, cast_to=str, required=True)
        super().__init__(name, path, **kwargs)

    def post_task_hook(self):
        self.git_add(self.param)

    @abstractmethod
    def task(self) -> Any:
        """
        Abstract method representing how a :func:`hammurabi.rules.base.Rule.task`
        must be parameterized. Any difference in the parameters will result in
        pylint/mypy errors.

        For more details please check :func:`hammurabi.rules.base.Rule.task`.
        """


class OwnerChanged(SingleAttributeRule):
    """
    Change the ownership of a file or directory.

    The new ownership of a file or directory can be set in three ways.
    To set only the user use ``new_value="username"``. To set only the
    group use ``new_value=":group_name"`` (please note the colon ``:``).
    It is also possible to set both username and group at the same time
    by using ``new_value="username:group_name"``.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, OwnerChanged
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         OwnerChanged(
        >>>             name="Change ownership of nginx config",
        >>>             path=Path("./nginx.conf"),
        >>>             new_value="www:web_admin"
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def task(self) -> Path:
        """
        Change the ownership of the given file or directory.
        None of the new username or group name can contain colons,
        otherwise only the first two colon separated values will be
        used as username and group name.

        :return: Return the input path as an output
        :rtype: Path
        """

        user, group = map(lambda x: x.strip(), self.new_value.partition(":")[::2])

        logging.debug('Changing owner of "%s" to "%s"', self.param, self.new_value)
        shutil.chown(str(self.param), user=user or None, group=group or None)

        return self.param


class ModeChanged(SingleAttributeRule):
    """
    Change the mode of a file or directory.

    Supported modes:

    +-------------------+-------------------------------------+
    | Config option     | Description                         |
    +===================+=====================================+
    | stat.S_ISUID      | Set user ID on execution.           |
    +-------------------+-------------------------------------+
    | stat.S_ISGID      | Set group ID on execution.          |
    +-------------------+-------------------------------------+
    | stat.S_ENFMT      | Record locking enforced.            |
    +-------------------+-------------------------------------+
    | stat.S_ISVTX      | Save text image after execution.    |
    +-------------------+-------------------------------------+
    | stat.S_IREAD      | Read by owner.                      |
    +-------------------+-------------------------------------+
    | stat.S_IWRITE     | Write by owner.                     |
    +-------------------+-------------------------------------+
    | stat.S_IEXEC      | Execute by owner.                   |
    +-------------------+-------------------------------------+
    | stat.S_IRWXU      | Read, write, and execute by owner.  |
    +-------------------+-------------------------------------+
    | stat.S_IRUSR      | Read by owner.                      |
    +-------------------+-------------------------------------+
    | stat.S_IWUSR      | Write by owner.                     |
    +-------------------+-------------------------------------+
    | stat.S_IXUSR      | Execute by owner.                   |
    +-------------------+-------------------------------------+
    | stat.S_IRWXG      | Read, write, and execute by group.  |
    +-------------------+-------------------------------------+
    | stat.S_IRGRP      | Read by group.                      |
    +-------------------+-------------------------------------+
    | stat.S_IWGRP      | Write by group.                     |
    +-------------------+-------------------------------------+
    | stat.S_IXGRP      | Execute by group.                   |
    +-------------------+-------------------------------------+
    | stat.S_IRWXO      | Read, write, and execute by others. |
    +-------------------+-------------------------------------+
    | stat.S_IROTH      | Read by others.                     |
    +-------------------+-------------------------------------+
    | stat.S_IWOTH      | Write by others.                    |
    +-------------------+-------------------------------------+
    | stat.S_IXOTH      | Execute by others.                  |
    +-------------------+-------------------------------------+

    Example usage:

    .. code-block:: python

        >>> import stat
        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, ModeChanged
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         ModeChanged(
        >>>             name="Update script must be executable",
        >>>             path=Path("./scripts/update.sh"),
        >>>             new_value=stat.S_IXGRP | stat.S_IXGRP | stat.S_IXOTH
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        new_value: Optional[int] = None,
        **kwargs
    ) -> None:
        # Passing the new value and then re-defining it is ugly, but needed
        # because the super class already has a validation on the new_value
        # field. For the first time it will be casted to string and then in
        # this __init__ it will be casted to integer. This logic can be
        # changed when it became frustrating.
        super().__init__(name, path, new_value=str(new_value), **kwargs)
        self.new_value = self.validate(self.new_value, cast_to=int, required=True)

    def task(self) -> Path:
        """
        Change the mode of the given file or directory.

        :return: Return the input path as an output
        :rtype: Path
        """

        logging.debug('Changing mode of "%s" to "%s"', self.param, self.new_value)
        os.chmod(str(self.param), self.new_value)
        return self.param



"""
Notifications are responsible for letting the end users/owners that a change
happened on a git repository. Notifications describes where to send the
notification but not responsible for delivering it. For example, you can use
an email notification method, but the notification method is not responsible
for handling emails and delivering the message.
"""


from abc import ABC, abstractmethod
import logging
from typing import List, Optional

from hammurabi.config import config
from hammurabi.exceptions import NotificationSendError
from hammurabi.helpers import full_strip


class Notification(ABC):
    """
    A ``git push`` notification which serves as a base for different kind
    of notifications like Slack or E-mail notification.
    """

    def __init__(self, recipients: List[str], message_template: str) -> None:
        self.recipients = recipients
        self.message_template = full_strip(message_template)

    def __format_message(self, changes_link: Optional[str]) -> str:
        """
        Fill the message template with actual values, such as repository owner/name
        and link to changes.

        :param changes_link: Link to the list of changes
        :type changes_link: Optional[str]

        :return: Returns the formatted message will be sent to the recipients
        :rtype: str
        """

        return self.message_template.format(
            repository=config.settings.repository, changes_link=changes_link
        )

    @abstractmethod
    def notify(self, message: str, changes_link: Optional[str]) -> None:
        """
        Handle sending the desired message to the recipients.

        :param message: Message to send
        :type message: str

        :param changes_link: Link to the list of changes
        :type changes_link: Optional[str]

        :raise: ``NotificationSendError`` if the notification cannot be delivered
        """

    def send(self, changes_link: Optional[str]) -> None:
        """
        Notify the users/owners about a change on the git repository. In case change
        link is provided, the user will be able to go directly checking the changes.

        :param changes_link: Link to the list of changes
        :type changes_link: Optional[str]
        """

        message = self.__format_message(changes_link)
        recipients = ", ".join(self.recipients)

        try:
            logging.info("Sending notification to %s", recipients)
            self.notify(message, changes_link)
        except NotificationSendError as exc:
            logging.error("Sending notification to %s failed: %s", recipients, str(exc))



from typing import Any, Dict, List, Union


class SelectorMixin:  # pylint: disable=too-few-public-methods
    """
    This mixin contains the helper function to get a value from dict by
    a css selector like selector path. (``.example.path.to.key``)
    """

    @staticmethod
    def __normalize_key_path(key_path: Union[str, List[str]]) -> List[str]:
        """
        Normalize the key_path and make sure we return the list
        representation of it.

        :param key_path: Path to the key in a selector format
            (``.path.to.the.key`` or ``["path", "to", "the", "key"]``)
        :type key_path: Union[str, List[str]]

        :return: List representation of key type
        :rtype: List[str]
        """

        if isinstance(key_path, str):
            key_path = key_path.split(".")

        return list(filter(lambda key: key, key_path))

    def get_by_selector(
        self, data: Any, key_path: Union[str, List[str]]
    ) -> Dict[str, Any]:
        """
        Get a key's value by a selector and traverse the path.

        :param data: The loaded YAML data into dict
        :type data: :class:`hammurabi.rules.mixins.Any`

        :param key_path: Path to the key in a selector format
            (``.path.to.the.key`` or ``["path", "to", "the", "key"]``)
        :type key_path: Union[str, List[str]]

        :return: Return the value belonging to the selector
        :rtype: :class:`hammurabi.rules.mixins.Any`
        """

        if not data:
            return dict()

        key_path = self.__normalize_key_path(key_path)
        entry = data or dict()

        for item in key_path:
            entry = entry.get(item, None)
            if not entry:
                return dict()

        return entry

    def set_by_selector(
        self,
        loaded_data: Any,
        key_path: Union[str, List[str]],
        value: Union[None, list, dict, str, int, float],
        delete: bool = False,
    ) -> Any:
        """
        Set a value by the key selector and traverse the path.

        :param loaded_data: The loaded YAML data into dict
        :type loaded_data: :class:`hammurabi.rules.mixins.Any`

        :param key_path: Path to the key in a selector format
            (``.path.to.the.key`` or ``["path", "to", "the", "key"]``)
        :type key_path: Union[str, List[str]]

        :param value: The value set for the key
        :type value: Union[None, list, dict, str, int, float]

        :param delete: Indicate if the key should be deleted
        :type delete: bool

        :return: The modified YAML data
        :rtype: :class:`hammurabi.rules.mixins.Any`
        """

        key_path = self.__normalize_key_path(key_path)
        data: Dict[str, Any] = loaded_data or dict()

        entry = data

        for item in key_path[:-1]:
            current = entry.get(item)

            if current and not isinstance(current, dict):
                entry[item] = {}

            entry = entry.setdefault(item, {})

        if not delete:
            entry[key_path[-1]] = value

        return data



"""
Text module contains simple but powerful general file content manipulations.
Combined with other simple rules like :class:`hammurabi.rules.files.FileExists`
or :class:`hammurabi.rules.attributes.ModeChanged` almost anything can be
achieved. Although any file's content can be changed using these rules, for
common file formats like ``ini``, ``yaml`` or ``json`` dedicated rules are
created.
"""


import logging
from pathlib import Path
import re
from typing import List, Optional, Tuple

from hammurabi.rules.common import SinglePathRule


class LineExists(SinglePathRule):
    """
    Make sure that the given file contains the required line. This rule is
    capable for inserting the expected text before or after the unique target
    text respecting the indentation of its context.

    The default behaviour is to insert the required text exactly after the
    target line, and respect its indentation. Please note that ``text``and
    ``target`` parameters are required.

    Example usage:

    .. code-block:: python

            >>> from pathlib import Path
            >>> from hammurabi import Law, Pillar, LineExists, IsLineNotExists
            >>>
            >>> gunicorn_config = Path("./gunicorn.conf.py")
            >>> example_law = Law(
            >>>     name="Name of the law",
            >>>     description="Well detailed description what this law does.",
            >>>     rules=(
            >>>         LineExists(
            >>>             name="Extend gunicorn config",
            >>>             path=gunicorn_config,
            >>>             text="keepalive = 65",
            >>>             target=r"^bind.*",
            >>>             preconditions=[
            >>>                 IsLineNotExists(path=gunicorn_config, criteria=r"^keepalive.*")
            >>>             ]
            >>>         ),
            >>>     )
            >>> )
            >>>
            >>> pillar = Pillar()
            >>> pillar.register(example_law)

    .. note::

        The indentation of the target text will be extracted by a simple
        regular expression. If a more complex regexp is required, please
        inherit from this class.
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        text: Optional[str] = None,
        target: Optional[str] = None,
        position: int = 1,
        respect_indentation: bool = True,
        ensure_trailing_newline: bool = False,
        **kwargs,
    ) -> None:
        self.text = self.validate(text, required=True)
        self.target = re.compile(self.validate(target, required=True))
        self.position = position
        self.respect_indentation = respect_indentation

        self.indentation_pattern = re.compile(r"^\s+")
        self.ensure_trailing_newline = ensure_trailing_newline

        super().__init__(name, path, **kwargs)

    def __get_target_match(self, lines: List[str]) -> str:
        """
        Get the matching target from the content of the given file.
        In case the matching number of lines are more than one or no
        match found, an exception will be raised accordingly.

        :param lines: Content of the given file
        :type lines: List[str]

        :raises: ``LookupError`` if no matching line can be found for target

        :return: List of the matching line
        :rtype: str
        """

        target_match = list(filter(self.target.match, lines))

        if not target_match:
            raise LookupError(f'No matching line for "{self.target}"')

        return target_match.pop()

    def __get_lines_from_file(self) -> Tuple[List[str], bool]:
        """
        Get the lines from the given file. In case of the file is empty, then
        append the expected line.

        :return: Returns the parsed lines and an indicator if the file was empty
        :rtype: tuple
        """

        file_was_empty = False

        with self.param.open("r") as file:
            logging.debug('Reading from "%s"', str(self.param))
            lines: List[str] = file.read().splitlines()

            if self.ensure_trailing_newline and lines[-1].strip() != "":
                lines.append("")

        if not lines:
            logging.debug('Adding "%s" to "%s"', self.text, str(self.param))
            lines.append(self.text)
            file_was_empty = True

        return lines, file_was_empty

    def __write_content_to_file(self, lines: List[str]):
        """
        Write the extended content of the file back. When writing the lines it is
        important to watch out for the new line character at the end of every line.

        :param lines: The new content of the original file
        :type lines: List[str]
        """

        with self.param.open("w") as file:
            file.writelines((f"{line}\n" for line in lines))

    def __add_line(self, lines: List[str]) -> None:
        """
        Make sure that the expected line is added to the list
        of lines.

        :param lines: Lines read from the input file
        :type lines: List[str]
        """

        target_match = self.__get_target_match(lines)

        # Get the index of the element from the right
        target_match_index = len(lines) - lines[::-1].index(target_match) - 1

        insert_position = target_match_index + self.position

        logging.debug('Inserting "%s" to position "%d"', self.text, insert_position)

        indentation = self.indentation_pattern.match(lines[target_match_index])
        if self.respect_indentation and indentation:
            self.text = indentation.group() + self.text

        lines.insert(insert_position, self.text)

    def task(self) -> Path:
        """
        Make sure that the given file contains the required line. This rule is
        capable for inserting the expected rule before or after the unique target
        text respecting the indentation of its context.

        :raises: ``LookupError``

        :return: Returns the path of the modified file
        :rtype: Path
        """

        lines, file_was_empty = self.__get_lines_from_file()

        if not file_was_empty:
            self.__add_line(lines)

        self.__write_content_to_file(lines)

        return self.param


class LineNotExists(SinglePathRule):
    """
    Make sure that the given file not contains the specified line.

    Example usage:

    .. code-block:: python

            >>> from pathlib import Path
            >>> from hammurabi import Law, Pillar, LineNotExists
            >>>
            >>> example_law = Law(
            >>>     name="Name of the law",
            >>>     description="Well detailed description what this law does.",
            >>>     rules=(
            >>>         LineNotExists(
            >>>             name="Remove keepalive",
            >>>             path=Path("./gunicorn.conf.py"),
            >>>             text="keepalive = 65",
            >>>         ),
            >>>     )
            >>> )
            >>>
            >>> pillar = Pillar()
            >>> pillar.register(example_law)
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        text: Optional[str] = None,
        **kwargs,
    ) -> None:
        self.text = re.compile(self.validate(text, cast_to=str, required=True))

        super().__init__(name, path, **kwargs)

    def task(self) -> Path:
        """
        Make sure that the given file not contains the specified line.

        :return: Returns the path of the modified file
        :rtype: Path
        """

        with self.param.open("r") as file:
            lines = file.read().splitlines()

        new_lines = list(filter(lambda l: not self.text.match(l), lines))

        if new_lines != lines:
            with self.param.open("w") as file:
                file.writelines((f"{line}\n" for line in new_lines))

        return self.param


class LineReplaced(SinglePathRule):
    """
    Make sure that the given text is replaced in the given file.

    The default behaviour is to replace the required text with the
    exact same indentation that the target line has. This behaviour
    can be turned off by setting the ``respect_indentation`` parameter
    to False.  Please note that ``text`` and ``target`` parameters are
    required.

    Example usage:

    .. code-block:: python

            >>> from pathlib import Path
            >>> from hammurabi import Law, Pillar, LineReplaced
            >>>
            >>> example_law = Law(
            >>>     name="Name of the law",
            >>>     description="Well detailed description what this law does.",
            >>>     rules=(
            >>>         LineReplaced(
            >>>             name="Replace typo using regex",
            >>>             path=Path("./gunicorn.conf.py"),
            >>>             text="keepalive = 65",
            >>>             target=r"^kepalive.*",
            >>>         ),
            >>>     )
            >>> )
            >>>
            >>> pillar = Pillar()
            >>> pillar.register(example_law)

    .. note::

        The indentation of the target text will be extracted by a simple
        regular expression. If a more complex regexp is required, please
        inherit from this class.

    .. warning::

        This rule will replace all the matching lines in the given file.
        Make sure the given target regular expression is tested before
        the rule used against production code.
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        text: Optional[str] = None,
        target: Optional[str] = None,
        respect_indentation: bool = True,
        **kwargs,
    ) -> None:
        self.text = self.validate(text, required=True)
        self.target = re.compile(self.validate(target, required=True))
        self.respect_indentation = respect_indentation

        self.indentation_pattern = re.compile(r"^\s+")

        super().__init__(name, path, **kwargs)

    def __get_lines_from_file(self) -> Tuple[List[str], bool]:
        """
        Get the lines from the given file.

        :return: Returns the parsed lines and an indicator if the file was empty
        :rtype: tuple
        """

        with self.param.open("r") as file:
            logging.debug('Reading from "%s"', str(self.param))
            lines = file.read().splitlines()

        return lines, not lines

    def __write_content_to_file(self, lines: List[str]):
        """
        Write the extended content of the file back. When writing the lines it is
        important to watch out for the new line character at the end of every line.

        :param lines: The new content of the original file
        :type lines: List[str]
        """

        with self.param.open("w") as file:
            file.writelines((f"{line}\n" for line in lines))

    def __replace_line(self, lines: List[str], target: str):
        """
        Replace the target texts with the given text.

        :param lines: The new content of the original file
        :type lines: List[str]

        :param target: The matching target in the given file's content
        :type target: str
        """

        target_index = lines.index(target)

        indentation = self.indentation_pattern.match(lines[target_index])
        if self.respect_indentation and indentation:
            self.text = indentation.group() + self.text

        lines[target_index] = self.text

    def task(self) -> Path:
        """
        Make sure that the given text is replaced in the given file.

        :raises: ``LookupError`` if we can not decide or can not find what should be replaced
        :return: Returns the path of the modified file
        :rtype: Path
        """

        lines, _ = self.__get_lines_from_file()

        target_match = list(filter(self.target.match, lines))
        text = list(filter(lambda l: l.strip() == self.text, lines))

        if target_match and text:
            raise LookupError(f'Both "{self.target}" and "{self.text}" exists')

        if text:
            return self.param

        if not target_match:
            raise LookupError(f'No matching line for "{self.target}"')

        for target in target_match:
            self.__replace_line(lines, target)

        self.__write_content_to_file(lines)

        return self.param



"""
Ini module is an extension for text rules tailor made for .ini/.cfg files.
The main difference lies in the way it works. First, the .ini/.cfg file is
parsed, then the modifications are made on the already parsed file.
"""

from abc import abstractmethod
import logging
from pathlib import Path
from typing import Any, Iterable, Optional, Tuple

from configupdater import ConfigUpdater  # type: ignore
from configupdater.configupdater import Section  # type: ignore

from hammurabi.rules.common import SinglePathRule


class SingleConfigFileRule(SinglePathRule):
    """
    Extend :class:`hammurabi.rules.base.Rule` to handle parsed content
    manipulations on a single file.

    .. warning::

        This rule requires the ``ini`` extra to be installed.
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        section: Optional[str] = None,
        **kwargs,
    ) -> None:
        self.section = self.validate(section, required=True)
        self.updater = ConfigUpdater()

        super().__init__(name, path, **kwargs)

    def pre_task_hook(self) -> None:
        """
        Parse the configuration file for later use.
        """

        logging.debug('Parsing "%s" configuration file', self.param)
        self.updater.read(self.param)

    @abstractmethod
    def task(self) -> Any:
        """
        Abstract method representing how a :func:`hammurabi.rules.base.Rule.task`
        must be parameterized. Any difference in the parameters will result in
        pylint/mypy errors.

        For more details please check :func:`hammurabi.rules.base.Rule.task`.
        """


class SectionExists(SingleConfigFileRule):
    """
    Ensure that the given config section exists. If needed, the rule will create
    a config section with the given name, and optionally the specified options. In
    case options are set, the config options will be assigned to that config sections.

    Similarly to :mod:`hammurabi.rules.text.LineExists`, this rule is able to add a
    section before or after a target section. The limitation compared to ``LineExists``
    is that the ``SectionExists`` rule is only able to add the new entry exactly before
    or after its target.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, SectionExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         SectionExists(
        >>>             name="Ensure section exists",
        >>>             path=Path("./config.ini"),
        >>>             section="polling",
        >>>             target="add_after_me",
        >>>             options=(
        >>>                 ("interval", "2s"),
        >>>                 ("abort_on_error", True),
        >>>             ),
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    .. warning::

        This rule requires the ``ini`` extra to be installed.

    .. warning::

        When ``options`` parameter is set, make sure you are using an iterable tuple.
        The option keys must be strings, but there is no limitation for the value. It can
        be set to anything what the parser can handle. For more information on the parser,
        please visit the documentation of  configupdater_.

        .. _configupdater: https://configupdater.readthedocs.io/en/latest/
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        target: Optional[str] = None,
        options: Iterable[Tuple[str, Any]] = (),
        add_after: bool = True,
        **kwargs,
    ) -> None:
        self.target = target
        self.options = options
        self.add_after = add_after

        self.space = 1

        super().__init__(name, path, **kwargs)

    def __get_target(self) -> Optional[Section]:
        """
        Get the target of the insert. If the target is not
        specified directly add as the last section.
        """

        if not self.updater.sections():
            return None

        if self.updater.has_section(self.target):
            return self.updater[self.target]

        return self.updater.sections_blocks()[-1]

    def __add_section(self) -> None:
        """
        Add the desired section before or after the target section if exists.
        In case the target section not exists, so the file was empty, simply
        add the new section.
        """

        logging.debug('Adding section "%s"', self.section)

        target = self.__get_target()

        if target is not None and self.add_after:
            target.add_after.space(self.space).section(self.section)
        elif target is not None and not self.add_after:
            target.add_before.section(self.section)
        else:
            self.updater.add_section(self.section)

    def __add_options(self) -> None:
        """
        Add options to the given section.
        """

        for option, value in self.options:
            self.updater[self.section][option] = value

    def task(self) -> Path:
        """
        Ensure that the given config section exists. If needed, create a config section with
        the given name, and optionally the specified options.

        :return: Return the input path as an output
        :rtype: Path
        """

        if not self.updater.has_section(self.section):
            self.__add_section()

        self.__add_options()

        with self.param.open("w") as file:
            self.updater.write(file)

        return self.param


class SectionNotExists(SingleConfigFileRule):
    """
    Make sure that the given file not contains the specified line. When a section
    removed, all the options belonging to it will be removed too.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, SectionNotExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         SectionNotExists(
        >>>             name="Ensure section removed",
        >>>             path=Path("./config.ini"),
        >>>             section="invalid",
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    .. warning::

        This rule requires the ``ini`` extra to be installed.
    """

    def task(self) -> Path:
        """
        Remove the given section including its options from the config file.

        :return: Return the input path as an output
        :rtype: Path
        """

        if self.updater.has_section(self.section):

            logging.debug('Removing section "%s"', self.section)
            self.updater.remove_section(self.section)

            with self.param.open("w") as file:
                self.updater.write(file)

        return self.param


class SectionRenamed(SingleConfigFileRule):
    """
    Ensure that a section is renamed. None of its options will be changed.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, SectionRenamed
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         SectionRenamed(
        >>>             name="Ensure section renamed",
        >>>             path=Path("./config.ini"),
        >>>             section="polling",
        >>>             new_name="fetching",
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    .. warning::

        This rule requires the ``ini`` extra to be installed.
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        new_name: Optional[str] = None,
        **kwargs,
    ) -> None:
        self.new_name = self.validate(new_name, required=True)

        super().__init__(name, path, **kwargs)

    def task(self) -> Path:
        """
        Rename the given section to a new name. None of its options will be
        changed. In case a section can not be found, a ``LookupError`` exception
        will be raised to stop the execution. The execution must be stopped at
        this point, because if other rules depending on the rename they will fail
        otherwise.

        :raises: ``LookupError`` if we can not decide or can not find what should be renamed
        :return: Return the input path as an output
        :rtype: Path
        """

        has_old_section = self.updater.has_section(self.section)
        has_new_section = self.updater.has_section(self.new_name)

        if has_old_section and has_new_section:
            raise LookupError(f'Both "{self.section}" and "{self.new_name}" set')

        if has_new_section:
            return self.param

        if not has_old_section:
            raise LookupError(f'No matching section for "{self.section}"')

        logging.debug('Renaming "%s" to "%s"', self.section, self.new_name)
        self.updater[self.section].name = self.new_name

        with self.param.open("w") as file:
            self.updater.write(file)

        return self.param


class OptionsExist(SingleConfigFileRule):
    """
    Ensure that the given config option exists. If needed, the rule will create
    a config option with the given value. In case the ``force_value`` parameter is
    set to True, the original values will be replaced by the give ones.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, OptionsExist
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         OptionsExist(
        >>>             name="Ensure options are changed",
        >>>             path=Path("./config.ini"),
        >>>             section="fetching",
        >>>             options=(
        >>>                 ("interval", "2s"),
        >>>                 ("abort_on_error", True),
        >>>             ),
        >>>             force_value=True,
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    .. warning::

        This rule requires the ``ini`` extra to be installed.

    .. warning::

        When using the ``force_value`` parameter, please note that all the existing
        option values will be replaced by those set in ``options`` parameter.
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        options: Iterable[Tuple[str, Any]] = None,
        force_value: bool = False,
        **kwargs,
    ) -> None:
        self.options = self.validate(options, required=True)
        self.force_value = force_value

        super().__init__(name, path, **kwargs)

    def task(self) -> Path:
        """
        Remove one or more option from a section. In case a section can not be
        found, a ``LookupError`` exception will be raised to stop the execution.
        The execution must be stopped at this point, because if dependant rules
        will fail otherwise.

        :raises: ``LookupError`` raised if no section can be renamed
        :return: Return the input path as an output
        :rtype: Path
        """

        if not self.updater.has_section(self.section):
            raise LookupError(f'No matching section for "{self.section}"')

        for option, value in self.options:
            if not self.updater.has_option(self.section, option) or self.force_value:
                logging.debug('Adding option "%s" = "%s"', option, value)
                self.updater[self.section][option] = value

        with self.param.open("w") as file:
            self.updater.write(file)

        return self.param


class OptionsNotExist(SingleConfigFileRule):
    """
    Remove one or more option from a section.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, OptionsNotExist
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         OptionsNotExist(
        >>>             name="Ensure options are removed",
        >>>             path=Path("./config.ini"),
        >>>             section="invalid",
        >>>             options=(
        >>>                 "remove",
        >>>                 "me",
        >>>                 "please",
        >>>             )
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    .. warning::

        This rule requires the ``ini`` extra to be installed.
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        options: Iterable[str] = (),
        **kwargs,
    ) -> None:
        self.options = self.validate(options, required=True)

        super().__init__(name, path, **kwargs)

    def task(self) -> Path:
        """
        Remove one or more option from a section. In case a section can not be
        found, a ``LookupError`` exception will be raised to stop the execution.
        The execution must be stopped at this point, because if dependant rules
        will fail otherwise.

        :raises: ``LookupError`` raised if no section can be renamed
        :return: Return the input path as an output
        :rtype: Path
        """

        if not self.updater.has_section(self.section):
            raise LookupError(f'No matching section for "{self.section}"')

        for option in self.options:
            logging.debug('Removing option "%s"', option)
            self.updater.remove_option(self.section, option)

        with self.param.open("w") as file:
            self.updater.write(file)

        return self.param


class OptionRenamed(SingleConfigFileRule):
    """
    Ensure that an option of a section is renamed.

    Example usage:

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, OptionRenamed
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         OptionRenamed(
        >>>             name="Rename an option",
        >>>             path=Path("./config.ini"),
        >>>             section="my_section",
        >>>             option="typo",
        >>>             new_name="correct",
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    .. warning::

        This rule requires the ``ini`` extra to be installed.
    """

    def __init__(
        self,
        name: str,
        path: Optional[Path] = None,
        option: Optional[str] = None,
        new_name: Optional[str] = None,
        **kwargs,
    ) -> None:
        self.option = self.validate(option, required=True)
        self.new_name = self.validate(new_name, required=True)

        super().__init__(name, path, **kwargs)

    def task(self) -> Path:
        """
        Rename an option of a section. In case a section can not be
        found, a ``LookupError`` exception will be raised to stop the execution.
        The execution must be stopped at this point, because if dependant rules
        will fail otherwise.

        :raises: ``LookupError`` raised if no section found or both the old and new
                 option names are found
        :return: Return the input path as an output
        :rtype: Path
        """

        if not self.updater.has_section(self.section):
            raise LookupError(f'No matching section for "{self.section}"')

        has_old_option = self.updater[self.section].get(self.option)
        has_new_option = self.updater[self.section].get(self.new_name)

        if has_old_option and has_new_option:
            raise LookupError(f'Both "{self.option}" and "{self.new_name}" set')

        if has_new_option:
            return self.param

        if not has_old_option:
            raise LookupError(f'No matching option for "{self.section}"')

        logging.debug(
            'Replacing option "%s" with "%s"', str(self.option), self.new_name
        )

        self.updater[self.section][self.option].name = self.new_name

        with self.param.open("w") as file:
            self.updater.write(file)

        return self.param



def full_strip(value: str) -> str:
    """
    Strip every line.
    """

    text = []

    for line in value.splitlines():
        stripped_line = line.strip()

        if stripped_line:
            text.append(stripped_line)

    return "\n".join(text)



"""
This module contains the definition of Preconditions which are related
to directories.
"""

from pathlib import Path

from hammurabi.preconditions.base import Precondition


class IsDirectoryExists(Precondition):
    """
    Check if the given directory exists.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, Renamed, IsDirectoryExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         Renamed(
        >>>             name="Rename the dir if an other one exists",
        >>>             path=Path("old-name"),
        >>>             new_name="new-name",
        >>>             preconditions=[
        >>>                 IsDirectoryExists(path=Path("other-dir"))
        >>>             ]
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)

    :param path: Input directory's path
    :type path: Path
    """

    def __init__(self, path: Path, **kwargs) -> None:
        super().__init__(param=path, **kwargs)

    def task(self) -> bool:
        """
        Check if the given directory exists.

        :return: Returns True if the directory exists
        :rtype: bool
        """

        self.param: Path
        return self.param.exists() and self.param.is_dir()


class IsDirectoryNotExists(IsDirectoryExists):
    """
    Opposite of :class:`hammurabi.preconditions.directories.IsDirectoryExists`.
    """

    def task(self) -> bool:
        """
        Check if the given directory not exists.

        :return: Returns True if the directory not exists
        :rtype: bool
        """

        return not super().task()



"""
Directories module contains directory specific manipulation rules. Please
note that those rules which can be used for files and directories are
located in other modules like :mod:`hammurabi.rules.operations` or
:mod:`hammurabi.rules.attributes`.
"""

import logging
import os
from pathlib import Path
import shutil

from hammurabi.rules.common import SinglePathRule


class DirectoryExists(SinglePathRule):
    """
    Ensure that a directory exists. If the directory does not exists,
    make sure the directory is created.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, DirectoryExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         DirectoryExists(
        >>>             name="Create secrets directory",
        >>>             path=Path("./secrets")
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def task(self) -> Path:
        """
        Create the given directory if not exists.

        :return: Return the input path as an output
        :rtype: Path
        """

        logging.debug('Creating directory "%s" if not exists', str(self.param))
        self.param.mkdir()

        return self.param


class DirectoryNotExists(SinglePathRule):
    """
    Ensure that the given directory does not exists.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, DirectoryNotExists
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         DirectoryNotExists(
        >>>             name="Remove unnecessary directory",
        >>>             path=Path("./temp")
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def post_task_hook(self):
        """
        Remove the given directory from git index.
        """

        self.git_remove(self.param)

    def task(self) -> Path:
        """
        Remove the given directory.

        :return: Return the input path as an output
        :rtype: Path
        """

        if self.param.exists():
            logging.debug('Removing directory "%s"', str(self.param))
            shutil.rmtree(self.param)

        return self.param


class DirectoryEmptied(SinglePathRule):
    """
    Ensure that the given directory's content is removed. Please note the
    difference between emptying a directory and recreating it. The latter
    results in lost ACLs, permissions and modes.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, DirectoryEmptied
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         DirectoryEmptied(
        >>>             name="Empty results directory",
        >>>             path=Path("./test-results")
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar()
        >>> pillar.register(example_law)
    """

    def task(self) -> Path:
        """
        Iterate through the entries of the given directory and remove them.
        If an entry is a file simply remove it, otherwise remove the whole
        subdirectory and its content.

        :return: Return the input path as an output
        :rtype: Path
        """

        with os.scandir(self.param) as entries:
            for entry in map(Path, entries):
                if entry.is_file() or entry.is_symlink():
                    logging.debug('Removing file "%s"', str(entry))
                    entry.unlink()

                elif entry.is_dir():
                    logging.debug('Removing directory "%s"', str(entry))
                    shutil.rmtree(entry)

        return self.param



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hammurabi',
 'hammurabi.notifications',
 'hammurabi.preconditions',
 'hammurabi.reporters',
 'hammurabi.rules']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'github3.py>=1.3.0,<2.0.0',
 'gitpython>=3.0.5,<4.0.0',
 'pydantic>=1.4,<2.0',
 'toml>=0.10.0,<0.11.0']

extras_require = \
{'all': ['ruamel.yaml>=0.16.5,<0.17.0',
         'configupdater>=1.0.1,<2.0.0',
         'ujson>=1.35,<4.0',
         'jinja2>=2.11.1,<3.0.0',
         'slack-webhook>=1.0.3,<2.0.0'],
 'ini': ['configupdater>=1.0.1,<2.0.0'],
 'slack-notifications': ['slack-webhook>=1.0.3,<2.0.0'],
 'templating': ['jinja2>=2.11.1,<3.0.0'],
 'ujson': ['ujson>=1.35,<4.0'],
 'yaml': ['ruamel.yaml>=0.16.5,<0.17.0']}

entry_points = \
{'console_scripts': ['hammurabi = hammurabi.main:cli']}

setup_kwargs = {
    'name': 'hammurabi',
    'version': '0.9.0',
    'description': 'Hammurabi is an extensible CLI tool responsible for enforcing user-defined rules on a git repository.',
    'long_description': 'Hammurabi\n*********\n\n.. image:: https://img.shields.io/pypi/v/hammurabi.svg\n    :target: https://pypi.python.org/pypi/hammurabi\n    :alt: PyPi Package\n\n.. image:: https://travis-ci.org/gabor-boros/hammurabi.svg?branch=master\n    :target: https://travis-ci.org/gabor-boros/hammurabi\n    :alt: Build Status\n\n.. image:: https://readthedocs.org/projects/hammurabi/badge/?version=latest\n    :target: https://hammurabi.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n\n.. image:: https://api.codeclimate.com/v1/badges/bcebab7105dfd82f358b/maintainability\n   :target: https://codeclimate.com/github/gabor-boros/hammurabi/maintainability\n   :alt: Maintainability\n\n.. image:: https://api.codeclimate.com/v1/badges/bcebab7105dfd82f358b/test_coverage\n    :target: https://codeclimate.com/github/gabor-boros/hammurabi/test_coverage\n    :alt: Test Coverage\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/ambv/black\n    :alt: Black Formatted\n\n.. image:: https://bestpractices.coreinfrastructure.org/projects/3587/badge\n    :target: https://bestpractices.coreinfrastructure.org/projects/3587\n    :alt: CII Best Practices\n\n\nMass changes made easy.\n\nHammurabi is an extensible CLI tool responsible for enforcing user-defined rules\non a git repository.\n\nFeatures\n========\n\nHammurabi integrates well with both git and Github to make sure that the\nexecution happens on a separate branch and the committed changes are pushed\nto the target repository. After pushing to the target repository, a pull\nrequest will be opened.\n\nHammurabi supports several operations (Rules) by default. These Rules can do\n\n* file and directory operations like copy, move, create or delete\n* manipulation of attributes like ownership or access permissions change\n* file and directory manipulations\n* piped rule execution (output of a rule is the input of the next rule)\n* children rule execution (output of a rule is the input of the upcoming rules)\n* creating files from Jinja2 templates\n* send notification on git push\n\nSupported file formats:\n\n* ``plain text``\n* ``ini``\n* ``json``\n* ``yaml`` (basic, single document operations)\n\nUpcoming file format support:\n\n* ``toml``\n* ``hocon``\n\nInstallation\n============\n\nHammurabi can be installed by running ``pip install hammurabi`` and it requires\nPython 3.7.0+ to run. This is the preferred method to install Hammurabi, as it\nwill always install the most recent stable release. If you don\'t have `pip`_\ninstalled, this `Python installation guide`_ can guide\nyou through the process.\n\n.. _pip: https://pip.pypa.io\n.. _Python installation guide: http://docs.python-guide.org/en/latest/starting/installation/\n\nInstalling extras\n-----------------\n\nHammurabi tries to be as tiny as its possible, hence some rules are requiring extra\ndependencies to be installed. Please check the documentation of the Rules to know\nwhich dependency is required to use the specific rule.\n\nTo install hammurabi with an extra package run ``pip install hammurabi[<EXTRA>]``,\nwhere ``<EXTRA>`` is the name of the extra option. To install multiple extra packages\nlist the extra names separated by comma as described in `pip\'s examples`_ section point\nnumber six.\n\n+---------------------+--------------------------------------------+\n| Extra               | Description                                |\n+=====================+============================================+\n| all                 | alias to install all the extras available  |\n+---------------------+--------------------------------------------+\n| ini                 | needed for ini/cfg based rules             |\n+---------------------+--------------------------------------------+\n| ujson               | install if you need faster json manipulat  |\n+---------------------+--------------------------------------------+\n| yaml                | needed for yaml based rules                |\n+---------------------+--------------------------------------------+\n| templating          | needed for rules which are using templates |\n+---------------------+--------------------------------------------+\n| slack-notifications | needed for slack webhook notifications     |\n+---------------------+--------------------------------------------+\n\n.. _`pip\'s examples`: https://pip.pypa.io/en/stable/reference/pip_install/#examples\n\nConfiguration\n=============\n\nFor configuration instructions, please visit the documentation_ site.\n\n.. _documentation: https://hammurabi.readthedocs.io/en/latest/config.html\n\nCommand line options\n====================\n\n.. code-block:: bash\n\n    hammurabi [OPTIONS] COMMAND [ARGS]...\n\n    Hammurabi is an extensible CLI tool responsible for enforcing user-defined\n    rules on a git repository.\n\n    Find more information at: https://hammurabi.readthedocs.io/latest/\n\n    Options:\n    -c, --config PATH               Set the configuration file.  [default:\n                                    pyproject.toml]\n    --repository TEXT               Set the remote repository. Required format:\n                                    owner/repository\n    --github-token TEXT             Set github access token\n    --log-level [DEBUG|INFO|WARNING|ERROR]\n                                    Set logging level.\n    --help                          Show this message and exit.\n\n    Commands:\n    describe  Show details of a specific resource or group of resources.\n    enforce   Execute all registered Law.\n    get       Show a specific resource or group of resources.\n    version   Print Hammurabi version.\n\nUsage examples\n==============\n\nIn every case, make sure that you clone the target repository prior using Hammurabi.\nAfter cloning the repository, always set the current working directory to the target\'s\npath. Hammurabi will not clone the target repository or change its execution directory.\n\nEnforce registered laws\n-----------------------\n\n.. code-block:: bash\n\n    $ hammurabi enforce\n    [INFO]  2020-14-07 16:31 - Checkout branch "hammurabi"\n    [INFO]  2020-14-07 16:31 - Executing law "L001"\n    [INFO]  2020-14-07 16:31 - Running task for "configure file exists"\n    [INFO]  2020-14-07 16:31 - Rule "configure file exists" finished successfully\n    [INFO]  2020-14-07 16:31 - Running task for "Minimum clang version is set"\n    [INFO]  2020-14-07 16:31 - Rule "Minimum clang version is set" finished successfully\n    [INFO]  2020-14-07 16:31 - Running task for "Minimum icc version is set"\n    [INFO]  2020-14-07 16:31 - Rule "Minimum icc version is set" finished successfully\n    [INFO]  2020-14-07 16:31 - Running task for "Minimum lessc version is set"\n    [INFO]  2020-14-07 16:31 - Rule "Minimum lessc version is set" finished successfully\n    [INFO]  2020-14-07 16:31 - Running task for "Maximum lessc version is set"\n    [INFO]  2020-14-07 16:31 - Rule "Maximum lessc version is set" finished successfully\n    [INFO]  2020-14-07 16:31 - Pushing changes\n    [INFO]  2020-14-07 16:35 - Checking for opened pull request\n    [INFO]  2020-14-07 16:35 - Opening pull request\n\nListing available laws\n----------------------\n\n.. code-block:: bash\n\n    $ hammurabi get laws\n    - Gunicorn config set up properly\n\nGet info about a law by its name\n--------------------------------\n\n.. code-block:: bash\n\n    $ hammurabi get law "Gunicorn config set up properly"\n    Gunicorn config set up properly\n\n    Change the gunicorn configuration based on our learnings\n    described at: https://google.com/?q=gunicorn.\n\n    If the gunicorn configuration does not exist, create a\n    new one configuration file.\n\nGet all registered (root) rules\n-------------------------------\n\n.. code-block:: bash\n\n    $ hammurabi get rules\n    - Rule 1\n    - Rule 5\n\nGet a rule by its name\n----------------------\n\n.. code-block:: bash\n\n    $ hammurabi get rule "Rule 1"\n    Rule 1\n\n    Ensure that a file exists. If the file does not exists,\n    this :class:`hammurabi.rules.base.Rule` will create it.\n\n    Due to the file is already created by :func:`pre_task_hook`\n    there is no need to do anything just return the input parameter.\n\nDescribe a law by its name\n--------------------------\n\n.. code-block:: bash\n\n    $ hammurabi describe law "Gunicorn config set up properly"\n    Gunicorn config set up properly\n\n    Change the gunicorn configuration based on our learnings\n    described at: http://docs.gunicorn.org/en/latest/configure.html.\n\n    If the gunicorn configuration does not exist, create a\n    new one configuration file.\n\n    Rules:\n    --> Rule 1\n    --> Rule 2\n    --> Rule 3\n    --> Rule 4\n    --> Rule 5\n\nDescribe a rule by its name\n---------------------------\n\n.. code-block:: bash\n\n    $ hammurabi describe rule "Rule 1"\n    Rule 1\n\n    Ensure that a file exists. If the file does not exists,\n    this :class:`hammurabi.rules.base.Rule` will create it.\n\n    Due to the file is already created by :func:`pre_task_hook`\n    there is no need to do anything just return the input parameter.\n\n    Chain:\n    --> Rule 1\n    --> Rule 2\n    --> Rule 3\n    --> Rule 4\n\nGetting the execution order of laws and rules\n---------------------------------------------\n\n.. code-block:: bash\n\n    $ hammurabi get order\n    - Gunicorn config set up properly\n    --> Rule 1\n    --> Rule 2\n    --> Rule 3\n    --> Rule 4\n    --> Rule 5\n\nCustom Rules\n============\n\nAlthough the project aims to support as many general operations as it can,\nthe need for adding custom rules may arise.\n\nTo extend Hammurabi with custom rules, you will need to inherit a class\nfrom ``Rule`` and define its abstract methods.\n\nThe following example will show you how to create and use a custom rule.\nFor more reference please check how the existing rules are implemented.\n\n.. code-block:: python\n\n    # custom.py\n    import shutil\n    import logging\n    from hammurabi.mixins import GitMixin\n    from hammurabi.rules.base import Rule\n\n\n    class CustomOwnerChanged(Rule, GitMixin):\n        """\n        Change the ownership of a file or directory to <original user>:admin.\n        """\n\n        def __init__(self, name: str, path: Optional[Path] = None, **kwargs):\n            super().__init__(name, path, **kwargs)\n\n        def post_task_hook(self):\n            self.git_add(self.param)\n\n        def task(self) -> Path:\n            # Since ``Rule`` is setting its 2nd parameter to ``self.param``,\n            # we can use ``self.param`` to access the target file\'s path.\n            logging.debug(\'Changing group of "%s" to admin\', str(self.param))\n            shutil.chown(self.param, group="admin")\n            return self.param\n\nCommunity\n=========\n\nIf you need help or you would like to be part of the Hammurabi community, join us on discord_.\n\n.. _discord: https://discord.gg/dj8Myk5\n\nContributing\n============\n\nHurray, You reached this section, which means you are ready\nto contribute.\n\nPlease read our contibuting guideline_. This guideline will\nwalk you through how can you successfully contribute to\nHammurabi.\n\n.. _guideline: https://github.com/gabor-boros/hammurabi/blob/master/CONTRIBUTING.rst\n\nInstallation\n------------\n\nFor development you will need poetry_ and pre-commit_. After poetry\ninstalled, simply run `poetry install`. This command will both create\nthe virtualenv and install development dependencies for you.\n\n.. _poetry: https://python-poetry.org/docs/#installation\n.. _pre-commit: https://pre-commit.com/#install\n\n\nUseful make Commands\n--------------------\n\n+------------------+-------------------------------------+\n| Command          | Description                         |\n+==================+=====================================+\n| help             | Print available make commands       |\n+------------------+-------------------------------------+\n| clean            | Remove all artifacts                |\n+------------------+-------------------------------------+\n| clean-build      | Remove build artifacts              |\n+------------------+-------------------------------------+\n| clean-mypy       | Remove mypy artifacts               |\n+------------------+-------------------------------------+\n| clean-pyc        | Remove Python artifacts             |\n+------------------+-------------------------------------+\n| clean-test       | Remove test artifacts               |\n+------------------+-------------------------------------+\n| docs             | Generate Sphinx documentation       |\n+------------------+-------------------------------------+\n| format           | Run several formatters              |\n+------------------+-------------------------------------+\n| lint             | Run several linters after format    |\n+------------------+-------------------------------------+\n| test             | Run all tests with coverage         |\n+------------------+-------------------------------------+\n| test-unit        | Run unit tests with coverage        |\n+------------------+-------------------------------------+\n| test-integration | Run integration tests with coverage |\n+------------------+-------------------------------------+\n\nWhy Hammurabi?\n==============\n\nHammurabi was the sixth king in the Babylonian dynasty,\nwhich ruled in central Mesopotamia from c. 1894 to 1595 B.C.\n\nThe Code of Hammurabi was one of the earliest and most\ncomplete written legal codes and was proclaimed by the\nBabylonian king Hammurabi, who reigned from 1792 to 1750 B.C.\nHammurabi expanded the city-state of Babylon along the Euphrates\nRiver to unite all of southern Mesopotamia. The Hammurabi code\nof laws, a collection of 282 rules, established standards for\ncommercial interactions and set fines and punishments to meet\nthe requirements of justice. Hammurabi’s Code was carved onto\na massive, finger-shaped black stone stele (pillar) that was\nlooted by invaders and finally rediscovered in 1901.\n',
    'author': 'Gábor Boros',
    'author_email': 'gabor.brs@gmail.com',
    'maintainer': 'Gábor Boros',
    'maintainer_email': 'gabor.brs@gmail.com',
    'url': 'https://github.com/gabor-boros/hammurabi/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Send notification to a slack channel when Hammurabi creates/updates a pull request.
"""

import logging
from typing import Optional

from slack_webhook import Slack  # type: ignore

from hammurabi.exceptions import NotificationSendError
from hammurabi.notifications.base import Notification


class SlackNotification(Notification):
    """
    Send slack notification through Slack webhooks.

    Example usage:

    .. code-block:: python

        >>> from pathlib import Path
        >>> from hammurabi import Law, Pillar, Renamed, IsDirectoryExists, SlackNotification
        >>>
        >>> example_law = Law(
        >>>     name="Name of the law",
        >>>     description="Well detailed description what this law does.",
        >>>     rules=(
        >>>         Renamed(
        >>>             name="Rename the dir if an other one exists",
        >>>             path=Path("old-name"),
        >>>             new_name="new-name",
        >>>             preconditions=[
        >>>                 IsDirectoryExists(path=Path("other-dir"))
        >>>             ]
        >>>         ),
        >>>     )
        >>> )
        >>>
        >>> pillar = Pillar(notifications=[
        >>>     SlackNotification(
        >>>         recipients=["https://slack.webhook.url"],
        >>>         message_template="Dear team, the {repository} has new update.",
        >>>     )
        >>> ])
        >>> pillar.register(example_law)

    .. warning::

        This notification requires the ``slack-notifications`` extra to be installed.
    """

    def notify(self, message: str, changes_link: Optional[str]) -> None:
        """
        Handle notification send through Slack webhooks.

        :param message: Message to send
        :type message: str

        :param changes_link: Link to the list of changes
        :type changes_link: Optional[str]
        """

        try:
            # Every recipient is a slack webhook URL
            for hook in self.recipients:
                logging.debug('Sending to hook "%s"', hook)
                Slack(url=hook).post(text=message)
                logging.debug("Notification sent")
        except Exception as exc:
            raise NotificationSendError(str(exc))
